from abc import ABC

from aws_cdk.aws_dynamodb import Attribute, Table, BillingMode, AttributeType, TableEncryption
from aws_cdk.core import Stack, RemovalPolicy


class ApiKeysDatabase(Table, ABC):
    def __init__(self, scope: Stack, table_name: str) -> None:
        self.__scope = scope
        self.__table_name = table_name

        super().__init__(
            scope=scope,
            id='ApiKeysDatabase',
            table_name=table_name,
            partition_key=Attribute(name='ApiKey', type=AttributeType.STRING),
            removal_policy=RemovalPolicy.DESTROY,
            point_in_time_recovery=True,
            billing_mode=BillingMode.PAY_PER_REQUEST,
            encryption=TableEncryption.DEFAULT,
        )

    @property
    def table_name(self) -> str:
        """
        Table name property.

        Overrides original parent method. Calling this method will not create a dependency to this resource.

        :return: Table name.
        """
        return self.__table_name

    @property
    def table_arn(self) -> str:
        """
        Table ARN property.

        Overrides original parent method. Calling this method will not create a dependency to this resource.

        :return: Table ARN.
        """
        return f'arn:aws:dynamodb:{self.__scope.region}:{self.__scope.account}:table/{self.__table_name}'

    @property
    def region(self) -> str:
        return self.__scope.region



from abc import abstractmethod, ABC

from aws_cdk.aws_iam import PolicyStatement
from aws_cdk.aws_lambda import Function, Code, Runtime
from aws_cdk.aws_logs import RetentionDays
from aws_cdk.core import Duration, Stack

from b_cfn_custom_api_key_authorizer.api_keys_database import ApiKeysDatabase
from b_cfn_custom_api_key_authorizer_layer.authorizer_layer import AuthorizerLayer


class BaseFunction(ABC, Function):
    def __init__(
            self,
            scope: Stack,
            name: str,
            api_keys_database: ApiKeysDatabase,
            authorizer_layer: AuthorizerLayer,
            *args,
            **kwargs
    ) -> None:
        super().__init__(
            scope=scope,
            id=name,
            function_name=name,
            code=self.code(),
            handler='index.handler',
            runtime=Runtime.PYTHON_3_8,
            log_retention=RetentionDays.ONE_MONTH,
            memory_size=128,
            timeout=Duration.seconds(30),
            *args,
            **kwargs
        )

        self.add_layers(authorizer_layer)

        # We want the functions to be able to access and manage the api keys database.
        self.add_environment('API_KEYS_DATABASE_NAME', api_keys_database.table_name)
        self.add_to_role_policy(
            PolicyStatement(
                actions=[
                    # Read actions.
                    'dynamodb:GetItem',
                    'dynamodb:Scan',
                    'dynamodb:Query',
                    'dynamodb:DescribeTable',
                    # Write actions.
                    'dynamodb:PutItem',
                    'dynamodb:DeleteItem',
                    'dynamodb:UpdateItem',
                    'dynamodb:BatchWriteItem'
                ],
                resources=[api_keys_database.table_arn]
            ))

    @abstractmethod
    def code(self) -> Code:
        pass



import logging

# These imports come from a layer.
from api_keys_verification import ApiKeysVerification
from auth_exception import AuthException

logger = logging.getLogger(__name__)


def handler(event, context):
    api_key = event['ApiKey']
    api_secret = event['ApiSecret']

    try:
        ApiKeysVerification(api_key, api_secret).verify()
        logger.info(f'Authentication succeeded for api key: {api_key}.')
        return {'Valid': True}
    except AuthException as ex:
        logger.info(f'Authentication failed for api key: {api_key}. Message: {repr(ex)}.')
        return {'Valid': False}



import json
from typing import Any, Dict, Optional

from b_aws_testing_framework.credentials import Credentials


class LambdaInvoke:
    def __init__(self, function_name: str):
        self.__function_name = function_name

    def invoke(self, payload: Optional[Dict[str, Any]] = None) -> Dict[str, Any]:
        kwargs = dict(
            FunctionName=self.__function_name,
            InvocationType='RequestResponse'
        )

        if payload is not None:
            kwargs['Payload'] = json.dumps(payload).encode()

        response = Credentials().boto_session.client('lambda').invoke(**kwargs)

        return json.loads(response['Payload'].read())



import logging
import os

import boto3
from botocore.exceptions import ClientError

from auth_exception import AuthException
from api_secret_hash import ApiSecretHash

logger = logging.getLogger(__name__)


class ApiKeysVerification:
    """
    Class responsible for api key verification. The inspiration is taken from this example:
    https://github.com/awslabs/aws-support-tools/blob/master/Cognito/decode-verify-jwt/decode-verify-jwt.py
    """

    def __init__(self, api_key: str, api_secret: str):
        self.__api_key = api_key
        self.__api_secret = api_secret
        self.__api_key_database_name = os.environ.get('API_KEYS_DATABASE_NAME')

        if (not api_key) or (not isinstance(api_key, str)):
            raise AuthException('Api Key not provided.')

        if (not api_secret) or (not isinstance(api_secret, str)):
            raise AuthException('Api Secret not provided.')

        if not self.__api_key_database_name:
            raise ValueError('Database not configured.')

    def verify(self) -> None:
        """
        Verifies the provided api key. If api key is not valid
        an exception is thrown. If no exception is thrown - api key is valid.

        :return: No return.
        """
        try:
            dynamodb = boto3.resource('dynamodb')
            table = dynamodb.Table(self.__api_key_database_name)
            data = table.get_item(Key={'ApiKey': self.__api_key})
        except ClientError:
            raise AuthException('Could not retrieve api key from database.')

        data = data.get('Item', {})
        api_key = data.get('ApiKey')
        api_secret_hash = data.get('ApiSecretHash')

        if (not api_key) or (not isinstance(api_key, str)):
            raise AuthException('Database error.')

        if (not api_secret_hash) or (not isinstance(api_secret_hash, str)):
            raise AuthException('Database error.')

        if not api_key == self.__api_key:
            logger.info(
                f'Given api key ({self.__api_key}) '
                f'does not match database api key ({api_key}).'
            )

            raise AuthException('Invalid authentication.')

        if not api_secret_hash == ApiSecretHash.hash_api_secret(self.__api_secret):
            logger.info(
                f'Given api secret ({self.__api_secret[:3]}... hash) '
                f'does not match database api secret hash.'
            )

            raise AuthException('Invalid authentication.')



import json
import logging
import os

# These imports come from a layer.
from api_keys_verification import ApiKeysVerification
from auth_exception import AuthException

from policy_document import PolicyDocument

# Allow extensive logging even with other file and other layer levels.
root_logger = logging.getLogger()
root_logger.setLevel(logging.INFO)
for handler in root_logger.handlers:
    handler.setLevel(logging.INFO)
logger = logging.getLogger(__name__)


def handler(event, context):
    logger.info(f'Received event:\n{json.dumps(event)}.')

    # Custom authorizer resource specifies header/ApiKey attribute:
    # identity_source=['$request.header.ApiKey', '$request.header.ApiSecret'].
    # They are converted to lowercase by apigateway/lambda services.
    api_key = event.get('headers', {}).get('apikey')
    api_secret = event.get('headers', {}).get('apisecret')

    document = PolicyDocument(
        region=os.environ['AWS_REGION'],
        account_id=os.environ['AWS_ACCOUNT'],
        api_id=os.environ['AWS_API_ID'],
        api_key=api_key
    )

    logger.info('Attempting to verify api keys...')

    # Verify the authorization token.
    try:
        ApiKeysVerification(api_key, api_secret).verify()
        logger.info(f'Authentication succeeded for api key: {api_key}.')
        # Authorization was successful. Return "Allow".
        return document.create_policy_statement(allow=True)
    except AuthException as ex:
        # Log the error.
        logger.info(f'Authentication failed for api key: {api_key}. Message: {repr(ex)}.')
        # Authorization has failed. Return "Deny".
        return document.create_policy_statement(allow=False)



from b_cfn_custom_api_key_authorizer_test.integration.infrastructure.main_stack import MainStack
from b_cfn_custom_api_key_authorizer_test.integration.util.lambda_invoke import LambdaInvoke


def test_FUNCTION_validator_WITH_valid_api_keys_EXPECT_response_valid(api_keys):
    function = MainStack.get_output(MainStack.VALIDATOR_FUNCTION)
    api_key, api_secret = api_keys

    response = LambdaInvoke(function).invoke({
        'ApiKey': api_key,
        'ApiSecret': api_secret
    })

    assert response['Valid'] is True


def test_FUNCTION_validator_WITH_invalid_api_keys_EXPECT_response_invalid():
    function = MainStack.get_output(MainStack.VALIDATOR_FUNCTION)

    response = LambdaInvoke(function).invoke({
        'ApiKey': '123',
        'ApiSecret': '456'
    })

    assert response['Valid'] is False



from typing import Callable, Tuple

from pytest import fixture

from b_cfn_custom_api_key_authorizer_test.integration.infrastructure.main_stack import MainStack
from b_cfn_custom_api_key_authorizer_test.integration.util.lambda_invoke import LambdaInvoke


@fixture(scope='function')
def api_keys_function() -> Callable[..., Tuple[str, str]]:
    """
    Fixture that returns a function.
    The function saves an api key - api secret pair object in database and
    returns a tuple (ApiKey, ApiSecret).

    This fixture does automatic cleanup (deletes the pair in the database) after test run.

    :return: Returns a function that saves api keys in database
        and returns a tuple representation of api keys.
    """
    generator_function = MainStack.get_output(MainStack.GENERATOR_FUNCTION)
    deleter_function = MainStack.get_output(MainStack.DELETER_FUNCTION)

    api_keys = []

    def __create_api_keys() -> Tuple[str, str]:
        response = LambdaInvoke(generator_function).invoke()

        # Extract generated api key and api secret.
        api_key = response['ApiKey']
        api_secret = response['ApiSecret']

        api_keys.append(api_key)

        return api_key, api_secret

    yield __create_api_keys

    for key in api_keys:
        LambdaInvoke(deleter_function).invoke({
            'ApiKey': key
        })


@fixture(scope='function')
def api_keys(api_keys_function) -> Tuple[str, str]:
    """
    Fixture that saves an api key - api secret pair object in database and
    returns a tuple (ApiKey, ApiSecret).

    This fixture does automatic cleanup (deletes the pair in the database) after test run.

    :return: Returns a tuple of api key and secret.
    """

    return api_keys_function()



import hashlib


class ApiSecretHash:
    @staticmethod
    def hash_api_secret(plain_text_api_secret: str) -> str:
        secret_hash = hashlib.sha256()
        secret_hash.update(plain_text_api_secret.encode('UTF-8'))
        secret_hash = secret_hash.hexdigest()
        return secret_hash



import os

from b_aws_cdk_parallel.deployment_executor import DeploymentExecutor
from b_aws_cdk_parallel.deployment_type import DeploymentType
from b_aws_testing_framework.credentials import Credentials
from b_aws_testing_framework.tools.cdk_testing.cdk_tool_config import CdkToolConfig
from b_aws_testing_framework.tools.cdk_testing.testing_manager import TestingManager

GLOBAL_PREFIX = os.environ.get('GLOBAL_PREFIX')
DO_NOT_CREATE_INFRASTRUCTURE = int(os.environ.get('DO_NOT_CREATE_INFRASTRUCTURE', 0))
DO_NOT_DESTROY_INFRASTRUCTURE = int(os.environ.get('DO_NOT_DESTROY_INFRASTRUCTURE', 0))

CDK_PATH = f'{os.path.dirname(os.path.abspath(__file__))}'
MANAGER = TestingManager(Credentials(), CdkToolConfig(CDK_PATH, destroy_before_preparing=False))
if GLOBAL_PREFIX: MANAGER.set_global_prefix(f'B{GLOBAL_PREFIX[:10]}')

# Import all fixtures.
# noinspection PyUnresolvedReferences
from .fixtures import *


def pytest_sessionstart(session):
    inf_create()


def pytest_sessionfinish(session, exitstatus):
    inf_destroy()


def inf_create():
    if DO_NOT_CREATE_INFRASTRUCTURE == 1:
        return

    def wrapper(cdk_config: CdkToolConfig):
        DeploymentExecutor(
            type=DeploymentType.DEPLOY,
            path=cdk_config.cdk_app_path,
            env=cdk_config.deployment_process_environment,
        ).run()

    MANAGER.prepare_infrastructure(wrapper)


def inf_destroy():
    if DO_NOT_DESTROY_INFRASTRUCTURE == 1:
        return

    def wrapper(cdk_config: CdkToolConfig):
        DeploymentExecutor(
            type=DeploymentType.DESTROY,
            path=cdk_config.cdk_app_path,
            env=cdk_config.deployment_process_environment,
        ).run()

    MANAGER.destroy_infrastructure(wrapper)



from aws_cdk.aws_apigatewayv2 import CfnApi
from aws_cdk.aws_lambda import Code, CfnPermission
from aws_cdk.core import Stack

from b_cfn_custom_api_key_authorizer.api_keys_database import ApiKeysDatabase
from b_cfn_custom_api_key_authorizer.functions.base_function import BaseFunction
from b_cfn_custom_api_key_authorizer_layer.authorizer_layer import AuthorizerLayer


class AuthorizerFunction(BaseFunction):
    def __init__(
            self,
            scope: Stack,
            name: str,
            parent_api: CfnApi,
            api_keys_database: ApiKeysDatabase,
            authorizer_layer: AuthorizerLayer
    ) -> None:
        super().__init__(
            scope=scope,
            name=name,
            api_keys_database=api_keys_database,
            authorizer_layer=authorizer_layer
        )

        # These environment variables are necessary for a lambda function to create
        # a policy document to allow/deny access. Read more here:
        # https://docs.aws.amazon.com/apigateway/latest/developerguide/api-gateway-lambda-authorizer-output.html
        self.add_environment('AWS_ACCOUNT', scope.account)
        self.add_environment('AWS_API_ID', parent_api.ref)

        CfnPermission(
            scope=scope,
            id='InvokePermission',
            action='lambda:InvokeFunction',
            function_name=self.function_name,
            principal='apigateway.amazonaws.com',
        )

    def code(self) -> Code:
        from .source import root
        return Code.from_asset(root)



from typing import Optional, List

from aws_cdk.aws_lambda import Runtime, LayerVersion, AssetCode
from aws_cdk.core import Stack


class AuthorizerLayer(LayerVersion):
    def __init__(self, scope: Stack) -> None:
        """
        Constructor.

        :param scope: CloudFormation stack.
        """
        super().__init__(
            scope=scope,
            id='AuthorizerLayer',
            code=AssetCode(self.get_source_path()),
            compatible_runtimes=self.runtimes()
        )

    @staticmethod
    def get_source_path() -> str:
        from . import layer_root
        return f'{layer_root}/source'

    @staticmethod
    def runtimes() -> Optional[List[Runtime]]:
        return [
            Runtime.PYTHON_3_8,
            Runtime.PYTHON_3_9
        ]



from aws_cdk.aws_lambda import Code
from aws_cdk.core import Stack

from b_cfn_custom_api_key_authorizer.api_keys_database import ApiKeysDatabase
from b_cfn_custom_api_key_authorizer.functions.base_function import BaseFunction
from b_cfn_custom_api_key_authorizer_layer.authorizer_layer import AuthorizerLayer


class ExistsFunction(BaseFunction):
    def __init__(
            self,
            scope: Stack,
            name: str,
            api_keys_database: ApiKeysDatabase,
            authorizer_layer: AuthorizerLayer
    ) -> None:
        super().__init__(
            scope=scope,
            name=name,
            api_keys_database=api_keys_database,
            authorizer_layer=authorizer_layer
        )

    def code(self) -> Code:
        from .source import root
        return Code.from_asset(root)



from b_cfn_custom_api_key_authorizer_test.integration.infrastructure.main_stack import MainStack
from b_cfn_custom_api_key_authorizer_test.integration.util.lambda_invoke import LambdaInvoke


def test_FUNCTION_exists_WITH_non_existent_api_keys_EXPECT_does_not_exist_response():
    function = MainStack.get_output(MainStack.EXISTS_FUNCTION)

    response = LambdaInvoke(function).invoke({
        'ApiKey': 'non-existent-api-key-123-abc'
    })

    assert response['Exists'] is False


def test_FUNCTION_exists_WITH_existing_api_keys_EXPECT_exists_response(api_keys):
    function = MainStack.get_output(MainStack.EXISTS_FUNCTION)
    api_key, _ = api_keys

    response = LambdaInvoke(function).invoke({
        'ApiKey': api_key
    })

    assert response['Exists'] is True



from b_cfn_custom_api_key_authorizer_test.integration.infrastructure.main_stack import MainStack
from b_cfn_custom_api_key_authorizer_test.integration.util.lambda_invoke import LambdaInvoke


def test_FUNCTION_generator_WITH_valid_configuration_EXPECT_generated():
    generator_function = MainStack.get_output(MainStack.GENERATOR_FUNCTION)
    exists_function = MainStack.get_output(MainStack.EXISTS_FUNCTION)

    response = LambdaInvoke(generator_function).invoke()
    api_key = response['ApiKey']

    response = LambdaInvoke(exists_function).invoke({'ApiKey': api_key})
    assert response['Exists'] is True



from b_cfn_custom_api_key_authorizer_test.integration.infrastructure.main_stack import MainStack
from b_cfn_custom_api_key_authorizer_test.integration.util.lambda_invoke import LambdaInvoke


def test_FUNCTION_deleter_WITH_non_existent_api_keys_EXPECT_no_error():
    function = MainStack.get_output(MainStack.DELETER_FUNCTION)
    LambdaInvoke(function).invoke({'ApiKey': 'non-existent-api-key-123-abc'})


def test_FUNCTION_deleter_WITH_existing_api_keys_EXPECT_deleted(api_keys):
    deleter_function = MainStack.get_output(MainStack.DELETER_FUNCTION)
    exists_function = MainStack.get_output(MainStack.EXISTS_FUNCTION)

    api_key, _ = api_keys

    response = LambdaInvoke(exists_function).invoke({'ApiKey': api_key})
    assert response['Exists'] is True

    LambdaInvoke(deleter_function).invoke({'ApiKey': api_key})

    response = LambdaInvoke(exists_function).invoke({'ApiKey': api_key})
    assert response['Exists'] is False



import urllib3

from b_cfn_custom_api_key_authorizer_test.integration.infrastructure.main_stack import MainStack


def test_FUNCTION_authorizer_WITH_valid_credentials_EXPECT_request_allowed_to_pass(api_keys) -> None:
    """
    Tests whether the authorizer allows the request to pass through, if the
    api key and api secret are valid.

    :return: No return.
    """
    api_key, api_secret = api_keys

    response = urllib3.PoolManager().request(
        method='GET',
        url=MainStack.get_output(MainStack.API_ENDPOINT),
        headers={
            'ApiKey': api_key,
            'ApiSecret': api_secret
        },
    )

    # Make sure response is successful.
    assert response.status == 200

    data = response.data
    data = data.decode()

    # Response from a dummy lambda function defined in the infrastructure main stack.
    assert data == 'Hello World!'


def test_authorizer_with_no_key_secret() -> None:
    """
    Tests whether the authorizer denies the request to pass through, if the
    api key and api secret are invalid.

    :return: No return.
    """
    response = urllib3.PoolManager().request(
        method='GET',
        url=MainStack.get_output(MainStack.API_ENDPOINT),
        headers={},
    )

    assert response.status == 401


def test_authorizer_with_non_existent_api_key_secret() -> None:
    """
    Tests whether the authorizer denies the request to pass through, if the
    api key and api secret are invalid.

    :return: No return.
    """
    response = urllib3.PoolManager().request(
        method='GET',
        url=MainStack.get_output(MainStack.API_ENDPOINT),
        headers={
            'ApiKey': '123',
            'ApiSecret': '123'
        },
    )

    assert response.status == 403


def test_authorizer_with_invalid_key_secret(api_keys) -> None:
    """
    Tests whether the authorizer denies the request to pass through, if the
    api key and api secret are invalid.

    :return: No return.
    """
    api_key, api_secret = api_keys

    response = urllib3.PoolManager().request(
        method='GET',
        url=MainStack.get_output(MainStack.API_ENDPOINT),
        headers={
            'ApiKey': api_key,
            'ApiSecret': '123'
        },
    )

    assert response.status == 403

    response = urllib3.PoolManager().request(
        method='GET',
        url=MainStack.get_output(MainStack.API_ENDPOINT),
        headers={
            'ApiKey': '123',
            'ApiSecret': api_secret
        },
    )

    assert response.status == 403



import logging
import os
import random
import string

import boto3
from botocore.exceptions import ClientError

# These imports come from a layer.
from api_secret_hash import ApiSecretHash

logger = logging.getLogger(__name__)


def handler(event, context):
    try:
        key = generate_api_key()
        secret = generate_api_secret()
        secret_hash = ApiSecretHash.hash_api_secret(secret)

        data = {
            'ApiKey': key,
            'ApiSecretHash': secret_hash
        }

        dynamodb = boto3.resource('dynamodb')
        table = dynamodb.Table(os.environ.get('API_KEYS_DATABASE_NAME'))
        table.put_item(Item=data)
    except ClientError as e:
        client_error_message = e.response['Error']['Message']
        logger.exception(f'Failed to call dynamodb table. {client_error_message}.')
        raise ValueError('Could not generate and save api keys.')

    return {
        'ApiKey': key,
        'ApiSecret': secret
    }


def generate_api_key() -> str:
    space = string.ascii_uppercase + string.digits
    return ''.join(random.choices(space, k=15))


def generate_api_secret() -> str:
    simple_punctuation = '!@#$%^&*()_+'
    space = string.ascii_letters + string.digits + simple_punctuation
    return ''.join(random.choices(space, k=30))



from aws_cdk.aws_apigatewayv2 import CfnApi, CfnStage
from aws_cdk.core import Construct
from b_aws_testing_framework.tools.cdk_testing.testing_stack import TestingStack

from b_cfn_custom_api_key_authorizer.custom_authorizer import ApiKeyCustomAuthorizer
from b_cfn_custom_api_key_authorizer_test.integration.infrastructure import authorized_endpoint_stack


class MainStack(TestingStack):
    API_ENDPOINT = 'ApiEndpoint'
    AUTHORIZER_FUNCTION = 'AuthorizerFunction'
    DELETER_FUNCTION = 'DeleterFunction'
    EXISTS_FUNCTION = 'ExistsFunction'
    GENERATOR_FUNCTION = 'GeneratorFunction'
    VALIDATOR_FUNCTION = 'ValidatorFunction'

    def __init__(self, scope: Construct) -> None:
        super().__init__(scope=scope)

        prefix = TestingStack.global_prefix()

        self.api = CfnApi(
            scope=self,
            id='Api',
            name=f'{prefix}Api',
            description='Sample description.',
            protocol_type='HTTP',
            cors_configuration=CfnApi.CorsProperty(
                allow_methods=['GET', 'PUT', 'POST', 'OPTIONS', 'DELETE'],
                allow_origins=['*'],
                allow_headers=[
                    'Content-Type',
                    'Authorization'
                ],
                max_age=300
            )
        )

        self.authorizer = ApiKeyCustomAuthorizer(
            scope=self,
            resource_name_prefix=prefix,
            api=self.api,
        )

        self.stage: CfnStage = CfnStage(
            scope=self,
            id='Stage',
            stage_name='test',
            api_id=self.api.ref,
            auto_deploy=True,
        )

        self.endpoint_stack = authorized_endpoint_stack.AuthorizedEndpointStack(self, self.api, self.authorizer)
        self.dummy_endpoint = f'{self.api.attr_api_endpoint}/{self.stage.stage_name}/{self.endpoint_stack.path}'

        self.add_output(self.API_ENDPOINT, value=self.dummy_endpoint)
        self.add_output(self.AUTHORIZER_FUNCTION, value=self.authorizer.authorizer_function.function_name)
        self.add_output(self.DELETER_FUNCTION, value=self.authorizer.deleter_function.function_name)
        self.add_output(self.EXISTS_FUNCTION, value=self.authorizer.exists_function.function_name)
        self.add_output(self.GENERATOR_FUNCTION, value=self.authorizer.generator_function.function_name)
        self.add_output(self.VALIDATOR_FUNCTION, value=self.authorizer.validator_function.function_name)



from aws_cdk.aws_apigatewayv2 import CfnAuthorizer, CfnApi
from aws_cdk.core import Stack

from b_cfn_custom_api_key_authorizer.api_keys_database import ApiKeysDatabase
from b_cfn_custom_api_key_authorizer.functions.authorizer.function import AuthorizerFunction
from b_cfn_custom_api_key_authorizer.functions.deleter.function import DeleterFunction
from b_cfn_custom_api_key_authorizer.functions.exists.function import ExistsFunction
from b_cfn_custom_api_key_authorizer.functions.generator.function import GeneratorFunction
from b_cfn_custom_api_key_authorizer.functions.validator.function import ValidatorFunction
from b_cfn_custom_api_key_authorizer_layer.authorizer_layer import AuthorizerLayer


class ApiKeyCustomAuthorizer(CfnAuthorizer):
    def __init__(
            self,
            scope: Stack,
            resource_name_prefix: str,
            api: CfnApi,
            cache_ttl: int = 60
    ) -> None:
        """
        Constructor.

        :param scope: CloudFormation stack.
        :param resource_name_prefix: Prefix string for all of the resources to be created.
            For example, if your prefix is "MyCoolProject", then the created api keys database
            will have a name of "MyCoolProjectApiKeysDatabase", api keys generator function
            will have a name of "MyCoolProjectApiKeysGeneratorFunction", and so on...
        :param api: Parent API for which we are creating the authorizer.
        :param cache_ttl: The TTL in seconds of cached authorizer results.
            If it equals 0, authorization caching is disabled.
            If it is greater than 0, API Gateway will cache authorizer responses.
            The maximum value is 3600, or 1 hour.
        """
        self.api_keys_database = ApiKeysDatabase(
            scope=scope,
            table_name=f'{resource_name_prefix}ApiKeysDatabase'
        )

        authorizer_layer = AuthorizerLayer(scope)

        base_kwargs = dict(
            scope=scope,
            api_keys_database=self.api_keys_database,
            authorizer_layer=authorizer_layer
        )

        # Authorizes requests.
        self.authorizer_function = AuthorizerFunction(
            **base_kwargs,
            name=f'{resource_name_prefix}ApiKeysAuthorizerFunction',
            parent_api=api,
        )

        # Deletes api keys.
        self.deleter_function = DeleterFunction(
            **base_kwargs,
            name=f'{resource_name_prefix}ApiKeysDeleterFunction',
        )

        # Checks whether api keys exist.
        self.exists_function = ExistsFunction(
            **base_kwargs,
            name=f'{resource_name_prefix}ApiKeysExistsFunction',
        )

        # Generates api keys.
        self.generator_function = GeneratorFunction(
            **base_kwargs,
            name=f'{resource_name_prefix}ApiKeysGeneratorFunction',
        )

        # Validates api keys.
        self.validator_function = ValidatorFunction(
            **base_kwargs,
            name=f'{resource_name_prefix}ApiKeysValidatorFunction',
        )

        # Constructed by reading this documentation:
        # https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-resource-apigatewayv2-authorizer.html
        super().__init__(
            scope=scope,
            id='ApiKeysCustomAuthorizer',
            name=f'{resource_name_prefix}ApiKeysCustomAuthorizer',
            api_id=api.ref,
            authorizer_payload_format_version='2.0',
            authorizer_result_ttl_in_seconds=cache_ttl,
            authorizer_type='REQUEST',
            authorizer_uri=(
                f'arn:aws:apigateway:{scope.region}:'
                f'lambda:path/2015-03-31/functions/arn:'
                f'aws:lambda:{scope.region}:{scope.account}:'
                f'function:{self.authorizer_function.function_name}/invocations'
            ),
            identity_source=[
                '$request.header.ApiKey',
                '$request.header.ApiSecret',
            ],
        )

    @property
    def authorization_type(self) -> str:
        """
        Property for authorization type when used with API Gateway service. Read more here:
        https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-resource-apigatewayv2-route.html#cfn-apigatewayv2-route-authorizationtype

        :return: Authorization type string.
        """
        return 'CUSTOM'



import logging
import os

import boto3
from botocore.exceptions import ClientError

logger = logging.getLogger(__name__)


def handler(event, context):
    api_key = event['ApiKey']

    try:
        dynamodb = boto3.resource('dynamodb')
        table = dynamodb.Table(os.environ.get('API_KEYS_DATABASE_NAME'))
        table.delete_item(Key={
            'ApiKey': api_key,
        })
    except ClientError as e:
        client_error_message = e.response['Error']['Message']
        logger.exception(f'Failed to call dynamodb table. {client_error_message}.')
        raise ValueError('Could not delete api keys.')
